@@ -41,6 +41,41 @@ start_server {tags {"slowlog"} overrides {slowlog-log-slower-than 1000000}} {
         assert_equal {foobar} [lindex $e 5]
     }
 
+    test {SLOWLOG - Rewritten commands are logged as their original command} {
+        r config set slowlog-log-slower-than 0
+
+        # Test rewriting client arguments
+        r sadd set a b c d e
+        r slowlog reset
+
+        # SPOP is rewritten as DEL when all keys are removed
+        r spop set 10
+        assert_equal {spop set 10} [lindex [lindex [r slowlog get] 0] 3]
+
+        # Test replacing client arguments
+        r slowlog reset
+
+        # GEOADD is replicated as ZADD
+        r geoadd cool-cities -122.33207 47.60621 Seattle
+        assert_equal {geoadd cool-cities -122.33207 47.60621 Seattle} [lindex [lindex [r slowlog get] 0] 3]
+
+        # Test replacing a single command argument
+        r set A 5
+        r slowlog reset
+        
+        # GETSET is replicated as SET
+        r getset a 5
+        assert_equal {getset a 5} [lindex [lindex [r slowlog get] 0] 3]
+
+        # INCRBYFLOAT calls rewrite multiple times, so it's a special case
+        r set A 0
+        r slowlog reset
+        
+        # INCRBYFLOAT is replicated as SET
+        r INCRBYFLOAT A 1.0
+        assert_equal {INCRBYFLOAT A 1.0} [lindex [lindex [r slowlog get] 0] 3]
+    }
+
     test {SLOWLOG - commands with too many arguments are trimmed} {
         r config set slowlog-log-slower-than 0
         r slowlog reset
@@ -70,7 +105,7 @@ start_server {tags {"slowlog"} overrides {slowlog-log-slower-than 1000000}} {
         assert_equal [lindex $e 3] {debug sleep 0.2}
     }
 
-    test {SLOWLOG - can clean older entires} {
+    test {SLOWLOG - can clean older entries} {
         r client setname lastentry_client
         r config set slowlog-max-len 1
         r debug sleep 0.2
@@ -78,4 +113,16 @@ start_server {tags {"slowlog"} overrides {slowlog-log-slower-than 1000000}} {
         set e [lindex [r slowlog get] 0]
         assert_equal {lastentry_client} [lindex $e 5]
     }
+
+    test {SLOWLOG - can be disabled} {
+        r config set slowlog-max-len 1
+        r config set slowlog-log-slower-than 1
+        r slowlog reset
+        r debug sleep 0.2
+        assert_equal [r slowlog len] 1
+        r config set slowlog-log-slower-than -1
+        r slowlog reset
+        r debug sleep 0.2
+        assert_equal [r slowlog len] 0
+    }
 }
